import time
from datetime import timedelta
from unittest.mock import patch

from ttlru_map import TTLMap
from ttlru_map._linked_list import DoubleLinkedListNode
from ttlru_map._ttl_map import _DictValue, _LinkedListValue


def test_update_by_ttl__empty_dict():
    d = TTLMap(ttl=timedelta(seconds=1000))
    d._update_by_ttl()


def test_update_by_ttl__last_item():
    ttl = timedelta(seconds=100)
    d = TTLMap(ttl=ttl)
    time_ = time.time() + ttl.total_seconds() + 1
    d._update_by_ttl(current_time=time_)
    assert d._ll_head is None
    assert d._ll_end is None
    assert d._dict == {}


def test_update_by_ttl__expired_head():
    ttl = timedelta(seconds=100)
    d = TTLMap(ttl=ttl)
    k1 = 1
    k2 = 2
    v1 = 1
    v2 = 2
    set_time_1 = 1
    set_time_2 = set_time_1 + ttl.total_seconds() + 1
    with patch("time.time", side_effect=[set_time_1, set_time_2]):
        d[k1] = v1
        d[k2] = v2
    d._update_by_ttl(current_time=set_time_2)
    node_2 = DoubleLinkedListNode(value=_LinkedListValue(time_=set_time_2, key=k2))
    assert d._ll_head == node_2
    assert d._ll_end == node_2
    assert d._dict == {k2: _DictValue(node=node_2, value=v2)}


def test_update_by_ttl__unlimited_ttl():
    size = 1000
    d = TTLMap(ttl=None, max_size=size)
    for i in range(size):
        d[i] = i
    d._update_by_ttl()
    assert len(d) == size



from datetime import timedelta

from ttlru_map import TTLMap


def test_pop_ll_node__middle_node():
    d = TTLMap(ttl=timedelta(seconds=1000))
    head = 1
    middle = 2
    end = 3

    d[head] = 1
    d[middle] = 2
    d[end] = 3

    head_node = d._dict[head].node
    middle_node = d._dict[middle].node
    end_node = d._dict[end].node

    assert head_node.next is middle_node
    assert middle_node.prev is head_node
    assert middle_node.next is end_node
    assert end_node.prev is middle_node
    assert d._ll_head is head_node
    assert d._ll_end is end_node

    d._pop_ll_node(middle_node)

    assert d._ll_head is head_node
    assert d._ll_end is end_node
    assert head_node.next is end_node
    assert end_node.prev is head_node
    assert middle_node.prev is None
    assert middle_node.next is None


def test_pop_ll_node__head_node():
    d = TTLMap(ttl=timedelta(seconds=1000))
    head = 1
    middle = 2
    end = 3

    d[head] = 1
    d[middle] = 2
    d[end] = 3

    head_node = d._dict[head].node
    middle_node = d._dict[middle].node
    end_node = d._dict[end].node

    assert head_node.next is middle_node
    assert middle_node.prev is head_node
    assert middle_node.next is end_node
    assert end_node.prev is middle_node
    assert d._ll_head is head_node
    assert d._ll_end is end_node

    d._pop_ll_node(head_node)

    assert d._ll_head is middle_node
    assert d._ll_end is end_node
    assert middle_node.prev is None
    assert middle_node.next is end_node
    assert head_node.prev is None
    assert head_node.next is None


def test_pop_ll_node__end_node():
    d = TTLMap(ttl=timedelta(seconds=1000))
    head = 1
    middle = 2
    end = 3

    d[head] = 1
    d[middle] = 2
    d[end] = 3

    head_node = d._dict[head].node
    middle_node = d._dict[middle].node
    end_node = d._dict[end].node

    assert head_node.next is middle_node
    assert middle_node.prev is head_node
    assert middle_node.next is end_node
    assert end_node.prev is middle_node
    assert d._ll_head is head_node
    assert d._ll_end is end_node

    d._pop_ll_node(end_node)

    assert d._ll_head is head_node
    assert d._ll_end is middle_node
    assert head_node.next is middle_node
    assert middle_node.prev is head_node
    assert end_node.prev is None
    assert end_node.next is None


def test_pop_ll_node__only_node():
    d = TTLMap(ttl=timedelta(seconds=1000))
    d[1] = 1
    assert d._ll_head is d._ll_end
    d._pop_ll_node(d._ll_head)
    assert d._ll_head is None
    assert d._ll_end is None



from __future__ import annotations

import time
from collections.abc import Hashable, MutableMapping
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Generic, TypeVar

from ttlru_map._exceptions import TTLMapInvalidConfigError
from ttlru_map._linked_list import DoubleLinkedListNode

if TYPE_CHECKING:
    from collections.abc import Iterator  # pragma: no cover
    from datetime import timedelta  # pragma: no cover

_TKey = TypeVar("_TKey", bound=Hashable)
_TValue = TypeVar("_TValue")


@dataclass(frozen=True)
class _LinkedListValue(Generic[_TKey]):
    __slots__ = ("key", "time_")

    time_: float
    key: _TKey

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(time_={self.time_}, key={self.key})"


@dataclass(frozen=True)
class _DictValue(Generic[_TKey, _TValue]):
    __slots__ = ("node", "value")
    node: DoubleLinkedListNode[_LinkedListValue[_TKey]]
    value: _TValue

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(node={self.node}, value={self.value})"


class TTLMap(MutableMapping[_TKey, _TValue]):
    """A dictionary that removes items after a certain time."""

    __slots__ = (
        "_dict",
        "_ll_end",
        "_ll_head",
        "_lock",
        "_max_size",
        "_ttl",
        "_update_ttl_on_get",
    )

    def __init__(
        self,
        *,
        ttl: timedelta | None,
        max_size: int | None = None,
        update_ttl_on_get: bool = False,
    ) -> None:
        """Initialize the dictionary.

        :param max_size: the maximum number of items in the dictionary. 0 means no limit.
        :param ttl: the time to live for each item in the dictionary.
        :param update_ttl_on_get: whether to update the time to live when getting an item.
        """
        self._validate_config(max_size, ttl, update_ttl_on_get)
        self._dict: dict[_TKey, _DictValue[_TKey, _TValue]] = {}
        self._ll_head: DoubleLinkedListNode[_LinkedListValue[_TKey]] | None = None
        self._ll_end: DoubleLinkedListNode[_LinkedListValue[_TKey]] | None = None
        self._max_size = max_size
        self._ttl = ttl
        self._update_ttl_on_get = update_ttl_on_get
        self._lock = Lock()

    @staticmethod
    def _validate_config(
        max_size: int | None,
        ttl: timedelta | None,
        update_ttl_on_get: bool,
    ) -> None:
        if max_size is None and ttl is None:
            msg = "max_size and ttl cannot be None at the same time."
            raise TTLMapInvalidConfigError(msg)
        if max_size is not None and max_size <= 0:
            msg = "max_size must be greater than 0."
            raise TTLMapInvalidConfigError(msg)
        if ttl is not None and ttl.total_seconds() <= 0:
            msg = "ttl must be greater than 0."
            raise TTLMapInvalidConfigError(msg)
        if ttl is None and update_ttl_on_get:
            msg = "update_ttl_on_get cannot be True when ttl is None."
            raise TTLMapInvalidConfigError(msg)

    def _update_by_ttl(self, current_time: float | None = None) -> None:
        """Remove items that have expired."""
        if self._ttl is None:
            return
        current_time = current_time if current_time is not None else time.time()
        while self._ll_head is not None:
            if self._ll_head.value.time_ + self._ttl.total_seconds() >= current_time:
                break
            del self._dict[self._ll_head.value.key]
            self._pop_ll_node(self._ll_head)

    def _update_by_size(self) -> None:
        """Remove the oldest items that exceed the maximum size."""
        if self._max_size is None:
            return
        while len(self._dict) > self._max_size and self._ll_head is not None:
            del self._dict[self._ll_head.value.key]
            self._pop_ll_node(self._ll_head)

    def _pop_ll_node(self, node: DoubleLinkedListNode[_LinkedListValue[_TKey]]) -> None:
        """Pop a node from the linked list."""
        if node is self._ll_head:
            self._ll_head = node.next
        if node is self._ll_end:
            self._ll_end = node.prev

        if node.next is not None:
            node.next.prev = node.prev
        if node.prev is not None:
            node.prev.next = node.next
        node.next = None
        node.prev = None

    def _put_node_to_end(self, node: DoubleLinkedListNode[_LinkedListValue[_TKey]]) -> None:
        """Put a node to the end of the linked list."""
        if self._ll_end is None:
            self._ll_head = node
            self._ll_end = node
        else:
            self._ll_end.next = node
            node.prev = self._ll_end
            self._ll_end = node

    def _setitem(self, __key: _TKey, __value: _TValue, time_: float, /) -> None:
        """Set an item in the dictionary and put it to the end of the linked list."""
        new_node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=__key))

        if (item := self._dict.get(__key, None)) is not None:
            self._pop_ll_node(item.node)
            self._put_node_to_end(new_node)
        else:
            self._put_node_to_end(new_node)

        self._dict[__key] = _DictValue(value=__value, node=new_node)

    def _delitem(self, item: _DictValue[_TKey, _TValue]) -> None:
        """Delete an item from the dictionary and the linked list."""
        del self._dict[item.node.value.key]
        self._pop_ll_node(item.node)

    def __setitem__(self, __key: _TKey, __value: _TValue, /) -> None:
        with self._lock:
            time_ = time.time()
            self._setitem(__key, __value, time_)
            self._update_by_ttl(current_time=time_)
            self._update_by_size()

    def __delitem__(self, __key: _TKey, /) -> None:
        with self._lock:
            item = self._dict[__key]
            self._delitem(item)
            self._update_by_ttl()

    def __getitem__(self, __key: _TKey, /) -> _TValue:
        with self._lock:
            time_ = time.time()
            self._update_by_ttl(current_time=time_)
            item = self._dict[__key].value
            if self._update_ttl_on_get:
                self._setitem(__key, item, time_)
            return item

    def __len__(self) -> int:
        with self._lock:
            self._update_by_ttl()
            return len(self._dict)

    def __iter__(self) -> Iterator[_TKey]:
        return iter(self._dict)



from datetime import timedelta

from ttlru_map import TTLMap


def test_update_by_size__unlimited_size():
    ttl = timedelta(seconds=1000)
    d = TTLMap(ttl=ttl, max_size=None)
    size = 1000
    for i in range(size):
        d[i] = i
    d._update_by_size()
    assert len(d) == size


def test_update_by_size__empty_dict():
    ttl = timedelta(seconds=1000)
    d = TTLMap(ttl=ttl, max_size=100)
    d._update_by_size()
    assert len(d) == 0


def test_update_by_size__dict_full():
    ttl = timedelta(seconds=1000)
    size = 3
    d = TTLMap(ttl=ttl, max_size=size)
    for i in range(size):
        d[i] = i
    assert len(d) == size
    d._max_size -= 1
    d._update_by_size()
    assert len(d) == size - 1
    assert d._dict == {i: d._dict[i] for i in range(1, size)}
    assert d._ll_head == d._dict[1].node
    assert d._ll_end == d._dict[size - 1].node


def test_update_by_size__rm_last_item():
    ttl = timedelta(seconds=1000)
    d = TTLMap(ttl=ttl, max_size=1)
    d[1] = 1
    assert len(d) == 1
    d._max_size = 0
    d._update_by_size()
    assert len(d) == 0
    assert d._ll_head is None
    assert d._ll_end is None



from datetime import timedelta

from ttlru_map import TTLMap
from ttlru_map._linked_list import DoubleLinkedListNode
from ttlru_map._ttl_map import _LinkedListValue


def test_put_node_to_end__empty_dict():
    d = TTLMap(ttl=timedelta(seconds=100))
    assert d._ll_head is None
    assert d._ll_end is None
    node = DoubleLinkedListNode(value=_LinkedListValue(time_=1, key=1))
    d._put_node_to_end(node=node)

    assert d._ll_head is node
    assert d._ll_end is node
    assert node.next is None
    assert node.prev is None


def test_put_node_to_end__not_empty_dict():
    d = TTLMap(ttl=timedelta(seconds=100))

    d[1] = 1
    d[2] = 2

    head_node = d._ll_head
    end_node = d._ll_end
    assert head_node is not None
    assert end_node is not None
    assert head_node is d._dict[1].node
    assert end_node is d._dict[2].node

    node = DoubleLinkedListNode(value=_LinkedListValue(time_=1, key=3))
    d._put_node_to_end(node=node)

    assert d._ll_head is head_node
    assert d._ll_end is node

    assert head_node.next is end_node
    assert end_node.prev is head_node
    assert end_node.next is node
    assert node.prev is end_node
    assert node.next is None



from datetime import timedelta
from unittest.mock import patch

from ttlru_map import TTLMap
from ttlru_map._linked_list import DoubleLinkedListNode
from ttlru_map._ttl_map import _DictValue, _LinkedListValue


def test_setitem__new_item():
    d = TTLMap(ttl=timedelta(seconds=1000))
    time_ = 100.0
    key = 1
    value = 1
    expected_node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    with (
        patch.object(TTLMap, "_pop_ll_node") as mock_pop_ll_node,
        patch.object(
            TTLMap,
            "_put_node_to_end",
        ) as mock_put_node_to_end,
    ):
        d._setitem(key, value, time_)
        mock_pop_ll_node.assert_not_called()
        mock_put_node_to_end.assert_called_once()
        assert mock_put_node_to_end.call_args[0][0] == expected_node
    assert d._dict == {key: _DictValue(value=value, node=expected_node)}


def test_setitem__existing_item():
    d = TTLMap(ttl=timedelta(seconds=1000))
    time_ = 100.0
    key = 1
    value = 1
    d[key] = value
    old_node = d._dict[key].node
    expected_node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    with (
        patch.object(TTLMap, "_pop_ll_node") as mock_pop_ll_node,
        patch.object(
            TTLMap,
            "_put_node_to_end",
        ) as mock_put_node_to_end,
    ):
        d._setitem(key, value, time_)
        mock_pop_ll_node.assert_called_once_with(old_node)
        mock_put_node_to_end.assert_called_once()
        assert mock_put_node_to_end.call_args[0][0] == expected_node
    assert d._dict == {key: _DictValue(value=value, node=expected_node)}



from datetime import timedelta

from ttlru_map import TTLMap


def test_len():
    d = TTLMap(ttl=timedelta(seconds=1000))
    assert len(d) == 0

    d[1] = 1
    assert len(d) == 1

    d[2] = 2
    assert len(d) == 2

    d[3] = 3
    assert len(d) == 3

    del d[1]
    assert len(d) == 2

    del d[2]
    assert len(d) == 1

    del d[3]
    assert len(d) == 0
